from libifstate.util import logger, IfStateLogging

import atexit
from copy import deepcopy
import logging
import multiprocessing as mp
import threading
import os
import re
from setproctitle import setproctitle
import signal
import subprocess
import sys

KEEPALIVED_HANDLE_STATES = ['backup', 'fault', 'master', 'stop', 'unknown']
KEEPALIVED_IGNORE_STATES = ['backup_priority', 'master_priority', 'master_rx_lower_pri']
KEEPALIVED_STATES = KEEPALIVED_HANDLE_STATES + KEEPALIVED_IGNORE_STATES

class VrrpFifoProcess():
    '''
    Process for vrrp group/instance configuration.
    '''
    def __init__(self, worker_args):
        self.worker_args = worker_args
        self.logger_extra = {'iface': f'{worker_args[-2]} "{worker_args[-1]}"'}
        self.state_queue = mp.Queue()
        self.worker_proc = None

        worker_io = threading.Thread(target=self.dequeue)
        worker_io.start()

    def vrrp_update(self, vrrp_state):
        self.state_queue.put(vrrp_state)

    def dequeue(self):
        while True:
            state = self.state_queue.get()

            # Should we terminate?
            if state is None:
                self.worker_proc.stdin.close()
                self.worker_proc.wait()
                return

            # Restart ifstate vrrp-worker if not alive alive?
            if self.worker_proc.poll() is not None:
                logger.warning("worker died", extra=self.logger_extra)
                self.start()

            logger.info(f'state => {state}', extra=self.logger_extra)
            self.worker_proc.stdin.write(f'{state}\n')
            self.worker_proc.stdin.flush()
        logger.warning("dequeue terminated")

    def start(self):
        logger.info("spawning worker", extra=self.logger_extra)
        self.worker_proc = subprocess.Popen(self.worker_args, stdin=subprocess.PIPE, stderr=sys.stderr, text=True)


class VrrpStates():
    '''
    Tracks processes and states for vrrp groups/instances.
    '''
    def __init__(self, ifs_config):
        self.ifs_config = ifs_config
        self.processes = {}
        self.states = {}
        self.pid_file = f"/run/libifstate/vrrp/{os.getpid()}.pid"
        self.cfg_file = f"/run/libifstate/vrrp/{os.getpid()}.cfg"
        self.worker_args = (
            sys.argv[0],
            '-c', self.cfg_file,
            'vrrp-worker')
        if logger.level == logging.DEBUG:
            self.worker_args.append('-v')

    def update(self, vrrp_type, vrrp_name, vrrp_state):
        '''
        Updates the state for a group/instance. A new VrrpFifoProcess is spawned on-demand
        if a group/instance is called the first time.
        '''
        key = (vrrp_type, vrrp_name)
        if not key in self.processes:
            worker_args = self.worker_args + key
            self.processes[key] = VrrpFifoProcess(worker_args)
            self.processes[key].start()

        self.states[key] = vrrp_state
        self.processes[key].vrrp_update(vrrp_state)

    def reconfigure(self, *argv):
        '''
        Reconfigure all known groups/instances using their last known state by spawning
        new worker Processes.
        '''

        # save new config to temp file
        try:
            self.ifs_config.load_config()
            self.ifs_config.dump_config(self.cfg_file)
        except Exception as ex:
            logger.error(f'failed to reload config: {ex}')
            return

        for key, state in self.states.items():
            worker_args = self.worker_args + key
            self.processes[key].vrrp_update(None)
            self.processes[key] = VrrpFifoProcess(worker_args)
            self.processes[key].start()
            self.processes[key].vrrp_update(self.states[key])

    def cleanup_run(self, *argv):
        """
        """
        for file in [self.pid_file, self.cfg_file]:
            try:
                os.unlink(file)
            except OSError as ex:
                if ex.errno != 2:
                    logger.warning(f"cannot cleanup {file}: {ex}")

def vrrp_fifo(args_fifo, ifs_config):
    vrrp_states = VrrpStates(ifs_config)

    signal.signal(signal.SIGHUP, vrrp_states.reconfigure)
    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    signal.signal(signal.SIGTERM, vrrp_states.cleanup_run)

    try:
        os.makedirs("/run/libifstate/vrrp", exist_ok=True)

        with open(vrrp_states.pid_file, "w", encoding="utf-8") as fh:
            fh.write(args_fifo)
        atexit.register(vrrp_states.cleanup_run)

        ifs_config.dump_config(vrrp_states.cfg_file)
    except IOError as err:
        logger.exception(f'failed to write pid file {vrrp_states.pid_file}: {err}')

    try:
        status_pattern = re.compile(
            r'(group|instance) "([^"]+)" (' + '|'.join(KEEPALIVED_STATES) + r')( \d+)?$', re.IGNORECASE)

        with open(args_fifo) as fifo:
            logger.debug("entering fifo loop...")
            for line in fifo:
                m = status_pattern.match(line.strip())
                if m:
                    vrrp_type = m.group(1)
                    vrrp_name = m.group(2)
                    vrrp_state = m.group(3)

                    # ignore priority changes w/o state changes
                    if vrrp_state.lower() in KEEPALIVED_IGNORE_STATES:
                        continue

                    vrrp_states.update(vrrp_type, vrrp_name, vrrp_state)
                else:
                    logger.warning(f'failed to parse fifo input: {line.strip()}')
                mp.active_children()
    finally:
        vrrp_states.cleanup_run()

def vrrp_worker(vrrp_type, vrrp_name, ifs_config):
    instance_title = "vrrp-{}-{}".format(vrrp_type, vrrp_name)
    setproctitle("ifstate-{}".format(instance_title))

    logger.info('worker is alive')
    for state in sys.stdin:
        vrrp_state = state.strip()

        ifstate = deepcopy(ifs_config.ifs)
        ifstate.apply(vrrp_type, vrrp_name, vrrp_state)
    logger.info('terminating')



from libifstate.util import logger
from libifstate.link.base import Link
from libifstate.exception import LinkCannotAdd, NetnsUnknown

class VethLink(Link):
    def __init__(self, ifstate, netns, name, link, ethtool, vrrp, brport):
        # use the bind_netns implementation to create the peer in the
        # target netns
        if 'peer_netns' in link:
            link['bind_netns'] = link['peer_netns']
            del(link['peer_netns'])

        super().__init__(ifstate, netns, name, link, ethtool, vrrp, brport)

    def create(self, do_apply, sysctl, excpts, oper="add"):
        '''
        Update the link registry for the created peer interface, too.
        '''
        result = super().create(do_apply, sysctl, excpts, oper)

        try:
            bind_netns = self.get_bind_netns()
        except NetnsUnknown as ex:
            excpts.add(oper, ex)
            return excpts

        peer_link = next(iter(bind_netns.ipr.get_links(ifname=self.settings['peer'])), None)
        if peer_link is None:
            excpts.add(oper, LinkCannotAdd(-1,"Failed to get peer ifIndex"))
            return excpts

        self.ifstate.link_registry.add_link(bind_netns, peer_link)

        return result

    def get_if_attr(self, key):
        '''
        Quirk to convert the 'peer' attribute from a ifindex value
        to the real ifname (netns aware).
        '''
        if key != "peer":
            return super().get_if_attr(key)

        peer = super().get_if_attr("link")

        if peer is None:
            return None

        bind_netns = self.get_bind_netns()
        lnk = next(iter(bind_netns.ipr.get_links(peer)), None)
        return lnk.get_attr("IFLA_IFNAME")

    def set_bind_state(self, state):
        '''
        Set the bind state for the peer, too.
        '''
        super().set_bind_state(state)

        bind_netns = self.get_bind_netns()
        peer_link = next(iter(bind_netns.ipr.get_links(ifname=self.settings['peer'])), None)

        fn = self.get_bind_fn(bind_netns.netns, peer_link['index'])
        try:
            with open(fn, 'wb') as fh:
                fh.write(self.netns.mount)
        except IOError as err:
            logger.warning('failed write `{}`: {}'.format(fn, err.args[1]))



from libifstate.util import logger
from libifstate.link.physical import PhysicalLink

class DsaLink(PhysicalLink):
    """
    Distributed Switch Architecture (DSA) user interface
    
    https://docs.kernel.org/networking/dsa/configuration.html
    """
    pass



from libifstate.util import logger, IfStateLogging
from libifstate.exception import ExceptionCollector, netlinkerror_classes

import errno
from pyroute2 import NetlinkError

class TC():
    ROOT_HANDLE = 0xFFFFFFFF
    INGRESS_HANDLE = 0xFFFF0000
    INGRESS_PARENT = 0xFFFFFFF1
    HMASK_MAJOR = 0xFFFF0000
    HMASK_MINOR = 0x0000FFFF

    def int2handle(h):
        maj = "{:x}".format((h & TC.HMASK_MAJOR) >> 16)
        mi = (h & TC.HMASK_MINOR)

        if mi == 0:
            mi = ""
        else:
            mi = "{:x}".format(mi)

        return ":".join((maj, mi))

    def handle2int(s):
        if isinstance(s, int):
            return s

        if s == "root":
            return TC.ROOT_HANDLE

        l = s.split(':')
        if l[1] == "":
            l[1] = "0"

        return int(l[0], 16) << 16 | int(l[1], 16)

    def __init__(self, netns, iface, tc):
        self.netns = netns
        self.iface = iface
        self.idx = None
        self.tc = tc

    def get_qdisc(self, netns_qdiscs, parent):
        for qdisc in netns_qdiscs:
            if qdisc['parent'] == parent:
                return qdisc

    def get_qchild(self, netns_qdiscs, parent, slot):
        for qdisc in netns_qdiscs:
            if qdisc['parent'] == parent | slot:
                return qdisc

    def apply_ingress(self, ingress, qdisc, excpts, do_apply):
        logger.debug('checking ingress qdisc', extra={'iface': self.iface})
        if not ingress:
            if qdisc:
                if do_apply:
                    opts = {
                        "index": self.idx,
                        "parent": TC.INGRESS_PARENT,
                    }
                    try:
                        self.netns.ipr.tc("del", **opts)
                    except Exception as err:
                        if not isinstance(err, netlinkerror_classes):
                            raise
                        logger.warning('removing ingress qdisc on {} failed: {}'.format(
                            self.iface, err.args[1]))
                        excpts.add('del', err, **opts)
                return True
        else:
            if not qdisc:
                if do_apply:
                    opts = {
                        "index": self.idx,
                        "kind": "ingress",
                    }
                    try:
                        self.netns.ipr.tc("add", **opts)
                    except Exception as err:
                        if not isinstance(err, netlinkerror_classes):
                            raise
                        logger.warning('adding ingress qdisc on {} failed: {}'.format(
                            self.iface, err.args[1]))
                        excpts.add('add', err, **opts)
                return True

        return False

    def apply_qtree(self, tc, qdisc, ipr_qdiscs, parent, excpts, do_apply, recreate=False):
        # if tc is None, we shall cleanup the qdisc tree
        # (remove the root qdisc to restore the interface's default)
        if tc is None:
            if do_apply:
                try:
                    self.netns.ipr.tc("del", index=self.idx, parent=parent)
                    return True
                except NetlinkError as ex:
                    # the native qdisc cannot be removed but only replaced,
                    # so we ignore any ENOENT
                    if ex.code != errno.ENOENT:
                        logger.exception('cannot remove qdisc')
            return False

        if qdisc is None:
            recreate = True

        if not recreate:
            logger.debug('  kind: {} => {}'.format(
                qdisc.get_attr("TCA_KIND"), tc["kind"]), extra={'iface': self.iface})
            if tc["kind"] != qdisc.get_attr("TCA_KIND"):
                recreate = True
            else:
                logger.debug('  handle: {} => {}'.format(
                    TC.int2handle(qdisc["handle"]), tc["handle"]), extra={'iface': self.iface})
                if TC.handle2int(tc["handle"]) != qdisc["handle"]:
                    recreate = True

        if recreate and do_apply:
            if qdisc is not None:
                try:
                    self.netns.ipr.tc("del", index=self.idx, parent=qdisc["parent"])
                except NetlinkError as ex:
                    # the native qdisc cannot be removed but only replaced,
                    # so we ignore any ENOENT
                    if ex.code != errno.ENOENT:
                        logger.exception('cannot remove qdisc')

        opts = {
            "index": self.idx,
            "parent": TC.int2handle(parent),
        }
        for k, v in tc.items():
            if k != "children":
                opts[k] = v

        if do_apply:
            if recreate:
                try:
                    self.netns.ipr.tc("add", **opts)
                except Exception as err:
                    if not isinstance(err, netlinkerror_classes):
                        raise
                    logger.warning('adding qdisc {} on {} failed: {}'.format(
                        opts.get("handle"), self.iface, err.args[1]))
                    excpts.add('add', err, **opts)
            else:
                # soft update for TCA_OPTIONS
                try:
                    self.netns.ipr.tc("change", **opts)
                except Exception as err:
                    if not isinstance(err, netlinkerror_classes):
                        raise
                    logger.warning('updating qdisc {} on {} failed: {}'.format(
                        opts.get("handle"), self.iface, err.args[1]))
                    excpts.add('change', err, **opts)

        changes = recreate
        if "children" in tc:
            for i in range(len(tc["children"])):
                changes = changes or self.apply_qtree(tc["children"][i], self.get_qchild(
                    ipr_qdiscs, qdisc["handle"], i+1), ipr_qdiscs, TC.handle2int(tc["handle"]) | i + 1, do_apply, recreate)

        return changes

    def apply_filter(self, tc, ipr_filters, excpts, do_apply):
        tc_filters = {}
        # assign prio numbers if missing
        for i in range(len(tc)):
            if not "prio" in tc[i]:
                tc[i]["prio"] = 0xc001 - len(tc) + i

            parent = TC.handle2int(tc[i].get("parent", 0))
            if not parent in tc_filters:
                tc_filters[parent] = {}

            tc_filters[parent][tc[i]["prio"]] = tc[i]

        changes = False
        # remove unreferenced filters
        removed = {}
        for ipr_filter in ipr_filters:
            prio = ipr_filter["info"] >> 16

            parent = TC.handle2int(ipr_filter.get("parent", 0))
            rm = not parent in tc_filters
            rm |= parent in tc_filters and \
                prio not in tc_filters[parent]
            if rm and prio not in removed.get(parent, []):
                changes = True
                if not parent in removed:
                    removed[parent] = []
                removed[parent].append(prio)
                if do_apply:
                    opts = {
                        "index": self.idx,
                        "info": ipr_filter["info"],
                        "parent": parent,
                    }
                    try:
                        self.netns.ipr.tc("del-filter", **opts)
                    except Exception as err:
                        if not isinstance(err, netlinkerror_classes):
                            raise
                        logger.warning('deleting filter #{} on {} failed: {}'.format(
                            prio, self.iface, err.args[1]))
                        excpts.add('del', err, **opts)

        if do_apply:
            for parent in tc_filters.keys():
                for tc_filter in tc_filters[parent].values():
                    tc_filter['index'] = self.idx
                    if "action" in tc_filter:
                        for action in tc_filter["action"]:
                            if action["kind"] == "mirred":
                                # get ifindex
                                action["ifindex"] = next(
                                    iter(self.netns.ipr.link_lookup(ifname=action["dev"])), None)

                                if self.idx == None:
                                    logger.warning("filter #{} references unknown interface {}".format(
                                        tc_filter["prio"], action["dev"]), extra={'iface': self.iface})
                                    continue
                    if "parent" in tc_filter:
                        tc_filter["parent"] = TC.handle2int(tc_filter["parent"])
                    try:
                        try:
                            self.netns.ipr.tc("replace-filter", **tc_filter)
                            # replace seems only to work if there is no filter
                            # => something has changed
                            changes = True
                        except Exception as err:
                            if not isinstance(err, netlinkerror_classes):
                                raise
                            # replace does not work, supress changes result
                            # for now
                            #changes = True
                            opts = {
                                "index": self.idx,
                                "info": tc_filter["prio"] << 16,
                                "parent": parent,
                            }
                            self.netns.ipr.tc("del-filter", **opts)
                            self.netns.ipr.tc("add-filter", **tc_filter)

                    except Exception as err:
                        if not isinstance(err, netlinkerror_classes):
                            raise
                        logger.warning('replace filter #{} on {} failed: {}'.format(
                            tc_filter['prio'], self.iface, err.args[1]))
                        excpts.add('replace', err, **tc_filter)

        return changes

    def apply(self, do_apply):
        excpts = ExceptionCollector(self.iface, self.netns)

        # get ifindex
        self.idx = next(iter(self.netns.ipr.link_lookup(ifname=self.iface)), None)

        if self.idx == None:
            logger.warning('link missing', extra={'iface': self.iface})
            return

        changes = []
        ipr_qdiscs = None

        # apply ingress qdics
        if "ingress" in self.tc:
            ipr_qdiscs = self.netns.ipr.get_qdiscs(index=self.idx)
            if self.apply_ingress(self.tc["ingress"],
                                  self.get_qdisc(
                                      ipr_qdiscs, TC.INGRESS_PARENT),
                                  excpts,
                                  do_apply):
                changes.append("ingress")

        # apply qdisc tree
        if "qdisc" in self.tc:
            if ipr_qdiscs is None:
                ipr_qdiscs = self.netns.ipr.get_qdiscs(index=self.idx)
            logger.debug('checking qdisc tree', extra={'iface': self.iface})
            if self.apply_qtree(
                    self.tc["qdisc"],
                    self.get_qdisc(ipr_qdiscs, TC.ROOT_HANDLE),
                    ipr_qdiscs,
                    TC.ROOT_HANDLE,
                    excpts,
                    do_apply):
                changes.append("qdisc")

        # apply filters
        if "filter" in self.tc:
            ipr_filters = self.netns.ipr.get_filters(
                index=self.idx) + self.netns.ipr.get_filters(index=self.idx, parent=TC.INGRESS_HANDLE)
            logger.debug('checking filters', extra={'iface': self.iface})
            if self.apply_filter(
                    self.tc["filter"],
                    ipr_filters,
                    excpts,
                    do_apply):
                changes.append("filter")

        if len(changes) > 0:
            logger.log_change('tc', 'change ({})'.format(", ".join(changes)))
        else:
            logger.log_ok('tc')

        return excpts



from libifstate.util import logger
from libifstate.link.base import Link
from libifstate.exception import LinkCannotAdd

from pwd import getpwnam
from grp import getgrnam

class TunLink(Link):
    def __init__(self, ifstate, netns, name, link, ethtool, vrrp, brport):
        if 'tun_owner' in link and isinstance(link['tun_owner'], str):
            link['tun_owner'] = getpwnam(link['tun_owner'])[2]

        if 'tun_group' in link and isinstance(link['tun_group'], str):
            link['tun_group'] = getgrnam(link['tun_group'])[2]

        super().__init__(ifstate, netns, name, link, ethtool, vrrp, brport)
        self.cap_create = bool(link.get('tun_persist'))
 
    def create(self, do_apply, sysctl, excpts, oper="add"):
        if not self.cap_create:
            logger.warning('unable to create non-persistent tuntap link', extra={'iface': self.settings.get('ifname'), 'netns': self.netns})
        else:
            super().create(do_apply, sysctl, excpts, oper)



from libifstate.util import logger, IfStateLogging
import pyroute2.netns

import os


class Sysctl():
    def __init__(self, netns):
        self.sysctls = {}
        self.globals = {}
        self.netns = netns

    def add(self, iface, sysctl):
        self.sysctls[iface] = sysctl

    def add_global(self, proto, sysctl):
        self.globals[proto] = sysctl

    def set_sysctl(self, iface_current, iface_config, family, key, val, do_apply):
        if self.netns.netns is not None:
            pyroute2.netns.pushns(self.netns.netns)

        log_str = "{}/{}".format(family, key)
        if self.netns.netns is not None:
            log_str += "[netns={}]".format(self.netns.netns)

        try:
            fn = '/proc/sys/net/{}/conf/{}/{}'.format(family, iface_current, key)
            try:
                with open(fn) as fh:
                    current = fh.readline().rstrip()
            except OSError as err:
                logger.warning('reading sysctl {}/{} failed: {}'.format(
                    family, key, err.args[1]))
                return
            if current == str(val):
                logger.debug('  %s/%s: %s == %s', family, key, current, val, extra={'iface': iface_config, 'netns': self.netns})
                logger.log_ok(log_str)
                return False
            else:
                logger.debug('  %s/%s: %s => %s', family, key, current, val, extra={'iface': iface_config, 'netns': self.netns})
                if do_apply:
                    try:
                        with open(fn, 'w') as fh:
                            fh.writelines([str(val)])
                    except OSError as err:
                        logger.warning('updating sysctl {}/{} failed: {}'.format(
                            family, key, err.args[1]))
                logger.log_change(log_str)
                return True
        finally:
            if self.netns.netns is not None:
                pyroute2.netns.popns()

    def apply(self, iface_current, do_apply, iface_config=None):
        # if None the interface has already the final name..
        if iface_config is None:
            iface_config = iface_current

        if not iface_config in self.sysctls:
            logger.debug("no sysctl settings", extra={'iface': iface_config})
            return

        logger.debug('checking sysctl', extra={'iface': iface_config})

        changes = False
        for family in self.sysctls[iface_config].keys():
            for key, val in self.sysctls[iface_config][family].items():
                changes = self.set_sysctl(iface_current, iface_config, family, key, val, do_apply) or changes

    def has_settings(self, iface):
        return iface in self.sysctls

    def apply_globals(self, do_apply):
        changes = []
        for proto, sysctl in self.globals.items():
            for key, val in sysctl.items():
                if self.set_sysctl('..', '..', proto, key, val, do_apply):
                    changes.append(proto)
        return changes

    def has_globals(self):
        return len(self.globals) > 0



from libifstate.util import logger, IfStateLogging
from libifstate.exception import netlinkerror_classes
from ipaddress import ip_address
from pyroute2.netlink.rtnl.ndmsg import NUD_NOARP, NUD_PERMANENT, NTF_SELF
from pyroute2.config import AF_BRIDGE
import pyroute2.netlink.rtnl.ndmsg

class FDB():
    def __init__(self, netns, iface, fdb):
        self.netns = netns
        self.iface = iface
        self.fdb = {}
        self.state_mask = NUD_NOARP|NUD_PERMANENT

        for entry in fdb:
            lladdr = entry['lladdr'].lower()
            _entry = {
                'lladdr': lladdr,
            }

            if 'port' not in entry:
                _entry['port'] = 8472
            else:
                _entry['port'] = entry['port']

            if 'dst' in entry:
                _entry['dst'] = str(ip_address(entry['dst']))

            if 'state' in entry:
                _entry['state'] = 0
                for name, value in pyroute2.netlink.rtnl.ndmsg.states.items():
                    if name in entry['state']:
                        _entry['state'] |= value

            if 'flags' in entry:
                _entry['flags'] = 0
                for name, value in pyroute2.netlink.rtnl.ndmsg.flags.items():
                    if name in entry['flags']:
                        _entry['flags'] |= value
            else:
                _entry['flags'] = NTF_SELF

            for opt in ['nhid', 'src_vni', 'vni']:
                if opt in entry:
                    _entry[opt] = entry[opt]

            if not lladdr in self.fdb:
                self.fdb[lladdr] = [_entry]
            else:
                self.fdb[lladdr].append(_entry)

    def get_kernel_fdb(self):
        # get fdb entries (NUD_NOARP|NUD_PERMANENT)
        fdb = {}
        for entry in self.netns.ipr.get_neighbours(ifindex=self.idx, family=AF_BRIDGE):
            state = entry.get('state')

            # look for permanent (local) or noarp (static) entries, only
            if not state & self.state_mask:
                continue

            lladdr = entry.get_attr('NDA_LLADDR')
            _entry = {
                'lladdr': lladdr,
                'state': state,
                'flags': entry.get('flags')
            }

            attr = entry.get_attr('NDA_DST')
            if attr is not None:
                _entry['dst'] = str(ip_address(attr))

            attr = entry.get_attr('NDA_PORT')
            if attr is None or attr == 0:
                _entry['port'] = 8472
            else:
                _entry['port'] = attr

            for opt in ['nhid', 'src_vni', 'vni']:
                attr = entry.get_attr(f"NDA_{opt.upper()}")
                if attr is not None:
                    _entry[opt] = attr

            if not lladdr in fdb:
                fdb[lladdr] = [_entry]
            else:
                fdb[lladdr].append(_entry)

        return fdb

    def apply(self, fdb_ignore_re, do_apply):
        logger.debug('getting fdb', extra={'iface': self.iface})

        # get ifindex and lladdr
        link = next(iter(self.netns.ipr.get_links(ifname=self.iface)), None)

        if link == None:
            logger.warning('link missing', extra={'iface': self.iface})
            return

        self.idx = link['index']
        self.lladdr = link.get_attr('IFLA_ADDRESS')

        # prepare default state for entries w/o state specified (depends on link type)
        default_state = NUD_PERMANENT

        linkinfo = link.get_attr('IFLA_LINKINFO')
        if linkinfo and linkinfo.get_attr('IFLA_INFO_KIND') in ['vxlan']:
            default_state |= NUD_NOARP

        # set default_state if missing
        for lladdr, entries in self.fdb.items():
            for entry in entries:
                if not 'state' in entry:
                    entry['state'] = default_state

        # cleanup orphan fdb entries
        ipr_entries = self.get_kernel_fdb()
        for lladdr, entries in ipr_entries.items():
            # ignore lladdr of the link
            if lladdr == self.lladdr:
                continue

            # skip ignored lladdr
            if fdb_ignore_re is not None and fdb_ignore_re.match(lladdr):
                continue

            for entry in entries:
                # check if fdb entry is already present
                if lladdr not in self.fdb or entry not in self.fdb[lladdr]:
                    logger.log_del('fdb', '- {}'.format(lladdr))

                    # prepare arguments
                    args = {
                        'ifindex': self.idx,
                        'family': AF_BRIDGE
                    }
                    args.update(entry)
                    logger.debug("bridge fdb del: {}".format(
                        " ".join("{}={}".format(k, v) for k, v in args.items())))

                    if do_apply:
                        try:
                            self.netns.ipr.fdb("del", **args)
                        except Exception as err:
                            if not isinstance(err, netlinkerror_classes):
                                raise
                            logger.warning('remove {} from fdb failed: {}'.format(
                                entry['lladdr'], err.args[1]))

        # configure fdb entries
        ipr_entries = self.get_kernel_fdb()
        for lladdr, entries in self.fdb.items():
            for entry in entries:
                # check if fdb entry is already present
                if lladdr in ipr_entries:
                    if entry in ipr_entries[lladdr]:
                        logger.log_ok('fdb', '= {}'.format(lladdr))
                        continue

                # fdb entry needs to be added
                logger.log_add('fdb', '+ {}'.format(lladdr))

                # prepare arguments
                args = {
                    'ifindex': self.idx,
                    'family': AF_BRIDGE
                }
                args.update(entry)
                logger.debug("bridge fdb append: {}".format(
                    " ".join("{}={}".format(k, v) for k, v in args.items())))

                if do_apply:
                    try:
                        self.netns.ipr.fdb("append", **args)
                    except Exception as err:
                        if not isinstance(err, netlinkerror_classes):
                            raise
                        logger.warning('add {} to fdb failed: {}'.format(
                            entry['lladdr'], err.args[1]))



from libifstate.util import logger, IfStateLogging
from libifstate.exception import RouteDuplicate, netlinkerror_classes
from ipaddress import ip_address, ip_interface, ip_network, IPv6Network
from pyroute2.netlink.rtnl.fibmsg import FR_ACT_VALUES
from pyroute2.netlink.rtnl import rt_type
import collections.abc
from glob import glob
import os
import re
import sys
import socket
from socket import AF_INET, AF_INET6


def route_matches(r1, r2, fields=('dst', 'priority', 'proto'), indent=None):
    return _matches(r1, r2, fields, indent)

def parse_route(route, implicit_defaults=True):
    rt = {}

    if implicit_defaults:
        rt['type'] = route.get('type', 'unicast')

    if 'to' in route:
        dst = ip_network(route['to'])
        rt['dst'] = dst.with_prefixlen
        has_to = True
    else:
        has_to = False

    for key, lookup in RT_LOOKUPS_DICT.items():
        try:
            if implicit_defaults:
                rt[key] = route.get(key, RT_LOOKUPS_DEFAULTS[key])
                rt[key] = lookup.lookup_id(rt[key])
            else:
                if key in route:
                    rt[key] = route[key]
                    rt[key] = lookup.lookup_id(rt[key])
        except KeyError as err:
            # mapping not available - catch exception and skip it
            logger.warning('ignoring unknown %s "%s"', key, rt[key],
                            extra={'iface': rt['dst']})
            rt[key] = RT_LOOKUPS_DEFAULTS[key]

    if 'type' in rt and type(rt['type']) == str:
        rt['type'] = rt_type[rt['type']]

    if 'dev' in route:
        rt['oif'] = route['dev']

    if 'via' in route:
        via = ip_address(route['via'])

        if not has_to or via.version == dst.version:
            rt['gateway'] = str(via)
        else:
            if via.version == 4:
                rt['via'] = {
                    'family': int(AF_INET),
                    'addr': str(via),
                }
            else:
                rt['via'] = {
                    'family': int(AF_INET6),
                    'addr': str(via),
                }

    if 'src' in route:
        rt['prefsrc'] = route['src']

    if 'preference' in route:
        rt['priority'] = route['preference']
    elif has_to and implicit_defaults:
        if isinstance(dst, IPv6Network):
            rt['priority'] = 1024
        else:
            rt['priority'] = 0

    if 'vrrp' in route:
        rt['_vrrp'] = route['vrrp']

    return rt

def parse_routes(routes, implicit_defaults=True):
    return [parse_route(route, implicit_defaults) for route in routes]

def parse_rule(rule, implicit_defaults=True):
    ru = {}

    if implicit_defaults:
        ru['table'] = RTLookups.tables.lookup_id(rule.get('table', 254))
        ru['protocol'] = RTLookups.protos.lookup_id(rule.get('proto', 0))
        ru['tos'] = rule.get('tos', 0)
    else:
        if 'table' in rule:
            ru['table'] = RTLookups.tables.lookup_id(rule['table'])
        if 'proto' in rule:
            ru['protocol'] = RTLookups.protos.lookup_id(rule['proto'])
        if 'tos' in rule:
            ru['tos'] = rule['tos']

    if 'action' in rule and type(rule['action']) == str:
        ru['action'] = {
            "to_tbl": "FR_ACT_TO_TBL",
            "unicast": "FR_ACT_UNICAST",
            "blackhole": "FR_ACT_BLACKHOLE",
            "unreachable": "FR_ACT_UNREACHABLE",
            "prohibit": "FR_ACT_PROHIBIT",
            "nat": "FR_ACT_NAT",
        }.get(rule['action'], rule['action'])
    elif implicit_defaults:
        ru['action'] = "FR_ACT_TO_TBL"

    if 'fwmark' in rule:
        ru['fwmark'] = rule['fwmark']

    if 'ipproto' in rule:
        if type(rule['ipproto']) == str:
            ru['ip_proto'] = socket.getprotobyname(rule['ipproto'])
        else:
            ru['ip_proto'] = rule['ipproto']

    if 'to' in rule:
        ru['dst'] = str(ip_network(rule['to']).network_address)
        ru['dst_len'] = ip_network(rule['to']).prefixlen

    if 'from' in rule:
        if  rule['from'] != "all":
            ru['src'] = str(ip_interface(rule['from']).ip)
            ru['src_len'] = ip_network(ip_interface(rule['from']).network).prefixlen
        else:
            ru['src'] = "0.0.0.0"
            ru['src_len'] = 0

    if 'iif' in rule:
        ru['iifname'] = rule['iif']

    if 'oif' in rule:
        ru['oifname'] = rule['oif']

    if 'priority' in rule:
        ru['priority'] = rule['priority']

    if 'vrrp' in rule:
        ru['_vrrp'] = rule['vrrp']

    if 'family' in rule:
        assert rule['family'] in ("AF_INET", "inet", 2, "AF_INET6", "inet6", 10), "rule['family'] doesn't have a valid value. This should have already been rejected by the schema validation."
        if rule['family'] in ("AF_INET", "inet", 2):
            ru['family'] = AF_INET
        elif rule['family'] in ("AF_INET6", "inet6", 10):
            ru['family'] = AF_INET6

    return ru

def parse_rules(rules, implicit_defaults=True):
    return [parse_rule(rule, implicit_defaults) for rule in rules]

def rule_matches(r1, r2, fields=('priority', 'iif', 'oif', 'dst', 'metric', 'protocol'), indent=None):
    return _matches(r1, r2, fields, indent)

def _matches(r1, r2, fields, indent):
    for fld in fields:
        if not indent is None:
            logger.debug("{}: {} - {}".format(fld, r1.get(fld), r2.get(fld)), extra={'iface': indent})
        if fld in r1 and fld in r2:
            if r1[fld] != r2[fld]:
                return False
        elif fld in r1 or fld in r2:
            return False
    return True

VRRP_MATCH_IGNORE = 0
VRRP_MATCH_DISABLE = 1
VRRP_MATCH_ENABLE =  2

def vrrp_match(obj, by_vrrp, vrrp_type, vrrp_name, vrrp_state):
    if by_vrrp:
        # test if object has a vrrp condition at all
        if not '_vrrp' in obj:
            return VRRP_MATCH_IGNORE

        # test if vrrp type and name matches
        if vrrp_type != obj['_vrrp']['type'] or vrrp_name != obj['_vrrp']['name']:
            return VRRP_MATCH_IGNORE

        # test if state is matching
        if vrrp_state in obj['_vrrp']['states']:
            return VRRP_MATCH_ENABLE
        else:
            return VRRP_MATCH_DISABLE
    else:
        # ignore vrrp object when not in vrrp action
        if '_vrrp' in obj:
            return VRRP_MATCH_IGNORE


class RTLookup():
    def __init__(self, name):
        self.name = name
        self.str2id = {}
        self.id2str = {}

        for basedir in ['/usr/share/iproute2', '/usr/lib/iproute2', '/etc/iproute2']:
            fn = os.path.join(basedir, name)
            try:
                with open(fn, 'r') as fp:
                    self._parse(fp)
            except IOError as err:
                pass

        for fn in glob(os.path.join('/etc/iproute2', "{}.d".format(name), "*.conf")):
            try:
                with open(fn, 'r') as fp:
                    self._parse(fp)
            except IOError as err:
                pass

    def _parse(self, fp):
        for line in fp:
            m = re.search(r'^(\d+)\s+(\S+)$', line)
            if m:
                self.str2id[m.group(2)] = int(m.group(1))
                self.id2str[int(m.group(1))] = m.group(2)

    def lookup_id(self, key):
        if type(key) == int or key.isdecimal():
            return int(key)

        return self.str2id[key]

    def lookup_str(self, key):
        return self.id2str.get(key, str(key))


class RTLookups():
    tables = RTLookup('rt_tables')
    realms = RTLookup('rt_realms')
    scopes = RTLookup('rt_scopes')
    protos = RTLookup('rt_protos')
    group = RTLookup('group')

RT_LOOKUPS_DICT = {
    'table': RTLookups.tables,
    'scope': RTLookups.scopes,
    'proto': RTLookups.protos,
    'realm': RTLookups.realms,
}

RT_LOOKUPS_DEFAULTS = {
    'table': 254,
    'scope': 0,
    'proto': 3,
    'realm': 0,
    'tos': 0,
}

class Tables(collections.abc.Mapping):
    def __init__(self, netns):
        self.netns = netns
        self.tables = {
            254: [],
        }
    def __getitem__(self, key):
        if not key in self.tables:
            raise KeyError()

        return self.tables[key]

    def __iter__(self):
        return self.tables.__iter__()

    def __len__(self):
        return len(self.tables)

    def parse_ignores(self, ignores):
        parsed = []
        for ignore in ignores:
            ignore = parse_route(ignore, False)

            for attr in ['dev', 'oif', 'iif']:
                if attr in ignore and type(ignore[attr]) == str:
                    idx = next(
                        iter(self.netns.ipr.link_lookup(ifname=ignore[attr])), None)
                    if idx is None:
                        logger.debug("{} '{}' is unknown, skipping ignore".format(attr, ignore[attr]))
                        continue
                    else:
                        ignore[attr] = idx
            parsed.append(ignore)

        return parsed

    def add(self, route):
        # convert config dict into IFLA keys and values
        rt = parse_route(route)

        # track route in the corresponding table
        if not rt['table'] in self.tables:
            self.tables[rt['table']] = []
        self.tables[rt['table']].append(rt)

    def show_routes(self, ignores):
        ignores = self.parse_ignores(ignores)

        routes = []
        for route in self.netns.ipr.get_routes(family=AF_INET) + self.netns.ipr.get_routes(family=AF_INET6):
            # skip routes from local table
            table = route.get_attr('RTA_TABLE')
            if table == 255:
                continue

            # skip ignored routes
            ignore = False
            for iroute in ignores:
                if route_matches(route, iroute, iroute.keys()):
                    ignore = True
                    break
            if ignore:
                continue

            if route['dst_len'] > 0:
                dst = ip_network(
                    '{}/{}'.format(route.get_attr('RTA_DST'), route['dst_len'])).with_prefixlen
            elif route['family'] == AF_INET:
                dst = ip_network('0.0.0.0/0').with_prefixlen
            elif route['family'] == AF_INET6:
                dst = ip_network('::/0').with_prefixlen

            rt = {
                'to': dst,
            }

            if table != 254:
                rt['table'] = RTLookups.tables.lookup_str(table),

            dev = route.get_attr('RTA_OIF')
            if dev:
                link = next(iter(self.netns.ipr.get_links(dev)), None)
                if link:
                    rt['dev'] = link.get_attr('IFLA_IFNAME', dev)
                else:
                    rt['dev'] = dev

            via = route.get_attr('RTA_GATEWAY')
            if via:
                rt['via'] = via
            else:
                via = route.get_attr('RTA_VIA')
                if via and 'addr' in via:
                    rt['via'] = via['addr']

            realm = route.get_attr('RTA_FLOW')
            if realm:
                rt['realm'] = RTLookups.realms.lookup_str(realm)

            if route['scope'] != 0:
                rt['scope'] = RTLookups.scopes.lookup_str(route['scope'])

            if route['proto'] != 3:
                rt['proto'] = RTLookups.protos.lookup_str(route['proto'])

            if route['tos'] != 0:
                rt['tos'] = route['tos']

            src = route.get_attr('RTA_PREFSRC')
            if src:
                rt['src'] = src

            rtype = route['type']
            if rtype != 1:
                rt['type'] = rt_type[rtype]

            priority = route.get_attr('RTA_PRIORITY')
            if priority:
                rt['preference'] = priority

            routes.append(rt)

        return routes

    def kernel_routes(self, table):
        routes = []
        for route in self.netns.ipr.get_routes(table=table, family=AF_INET) + self.netns.ipr.get_routes(table=table, family=AF_INET6):
            # ignore RTM_F_CLONED routes
            if route['flags'] & 512:
                continue

            if route['dst_len'] > 0:
                dst = ip_network(
                    '{}/{}'.format(route.get_attr('RTA_DST'), route['dst_len'])).with_prefixlen
            elif route['family'] == AF_INET:
                dst = ip_network('0.0.0.0/0').with_prefixlen
            elif route['family'] == AF_INET6:
                dst = ip_network('::/0').with_prefixlen

            rt = {
                'table': route.get_attr('RTA_TABLE'),
                'type': route['type'],
                'dst': dst,
                'oif': route.get_attr('RTA_OIF'),
                'scope': route['scope'],
                'proto': route['proto'],
                'realm': route.get_attr('RTA_FLOW', 0),
                'tos': route['tos'],
                'priority': 0
            }

            gateway = route.get_attr('RTA_GATEWAY')
            if not gateway is None:
                rt['gateway'] = str(ip_address(gateway))

            via = route.get_attr('RTA_VIA')
            if not via is None:
                rt['via'] = via

            metric = route.get_attr('RTA_PRIORITY')
            if not metric is None:
                rt['metric'] = metric

            pref = route.get_attr('RTA_PREF')
            if not pref is None:
                rt['pref'] = pref

            prefsrc = route.get_attr('RTA_PREFSRC')
            if not prefsrc is None:
                rt['prefsrc'] = prefsrc

            priority = route.get_attr('RTA_PRIORITY')
            if not priority is None:
                rt['priority'] = priority

            routes.append(rt)
        return routes

    def apply(self, ignores, do_apply, by_vrrp, vrrp_type, vrrp_name, vrrp_state):
        ignores = self.parse_ignores(ignores)

        for table, croutes in self.tables.items():
            log_str = RTLookups.tables.lookup_str(table)
            if self.netns.netns != None:
                log_str += "[netns={}]".format(self.netns.netns)

            kroutes = self.kernel_routes(table)

            for route in sorted(croutes, key=lambda x: [str(x.get('gateway', x.get('via', ''))), x['dst']]):
                if 'oif' in route and type(route['oif']) == str:
                    oif = next(
                        iter(self.netns.ipr.link_lookup(ifname=route['oif'])), None)
                    if oif is None:
                        if 'gateway' in route:
                            logger.log_warn(log_str, '! {}: dev {} is unknown'.format(route['dst'], route['oif']))
                            del route['oif']
                        else:
                            logger.log_err(log_str, '! {}: dev {} is unknown'.format(route['dst'], route['oif']))
                            continue
                    else:
                        route['oif'] = oif
                found = False
                identical = False
                matched = vrrp_match(route, by_vrrp, vrrp_type, vrrp_name, vrrp_state)
                for i, kroute in enumerate(kroutes):
                    if route_matches(route, kroute):
                        # ignore kernel routes due to vrrp_match
                        if matched == VRRP_MATCH_IGNORE:
                            del kroutes[i]
                            break

                        # remove kernel routes due to vrrp_match
                        if matched == VRRP_MATCH_DISABLE:
                            break

                        del kroutes[i]
                        found = True
                        if route_matches(
                            route,
                            kroute,
                            [key for key in route.keys() if key[0] != '_'],
                            indent=route['dst']
                        ):
                            identical = True
                            break

                if matched not in [VRRP_MATCH_IGNORE, VRRP_MATCH_DISABLE]:
                    if identical:
                        logger.log_ok(log_str, "= {}".format(route['dst']))
                    else:
                        if found:
                            logger.log_change(log_str, "~ {}".format(route['dst']))
                        else:
                            logger.log_add(log_str, "+ {}".format(route['dst']))

                        logger.debug("ip route replace: {}".format(
                            " ".join("{}={}".format(k, v) for k, v in route.items())))
                        try:
                            if do_apply:
                                self.netns.ipr.route('replace', **route)
                        except Exception as err:
                            if not isinstance(err, netlinkerror_classes):
                                raise
                            logger.warning('route setup {} failed: {}'.format(
                                route['dst'], err.args[1]))

            for route in kroutes:
                ignore = False
                for iroute in ignores:
                    if route_matches(route, iroute, iroute.keys()):
                        ignore = True
                        break
                if ignore:
                    continue

                logger.log_del(log_str, "- {}".format(route['dst']))
                try:
                    if do_apply:
                        self.netns.ipr.route('del', **route)
                except Exception as err:
                    if not isinstance(err, netlinkerror_classes):
                        raise
                    logger.warning('removing route {} failed: {}'.format(
                        route['dst'], err.args[1]))


class Rules():
    def __init__(self, netns):
        self.netns = netns
        self.rules = []

    def add(self, rule):
        # convert config dict into IFLA keys and values
        ru = parse_rule(rule)

        # track rule
        self.rules.append(ru)

    def kernel_rules(self):
        rules = []
        for rule in self.netns.ipr.get_rules(family=AF_INET) + self.netns.ipr.get_rules(family=AF_INET6):
            ru = {
                'action': FR_ACT_VALUES.get(rule['action']),
                'table': rule.get_attr('FRA_TABLE'),
                'protocol': rule.get_attr('FRA_PROTOCOL'),
                'priority': rule.get_attr('FRA_PRIORITY', 0),
                'family': rule['family'],
                'tos': rule['tos'],
            }

            if rule['dst_len'] > 0:
                ru['dst'] = rule.get_attr('FRA_DST')
                ru['dst_len'] = rule['dst_len']

            if rule['src_len'] > 0:
                ru['src'] = rule.get_attr('FRA_SRC')
                ru['src_len'] = rule['src_len']

            for field in ['iifname', 'oifname', 'fwmark', 'ip_proto']:
                value = rule.get_attr('FRA_{}'.format(field.upper()))
                if not value is None:
                    ru[field] = value

            rules.append(ru)
        return rules

    def show_rules(self, ignores):
        ignores = parse_rules(ignores, implicit_defaults=False)

        rules = []
        for rule in self.kernel_rules():
            # skip ignored routes
            ignore = False
            for irule in ignores:
                if rule_matches(rule, irule, irule.keys()):
                    ignore = True
                    break
            if ignore:
                continue

            rule['action'] = {
                "FR_ACT_TO_TBL": "to_tbl",
                "FR_ACT_UNICAST": "unicast",
                "FR_ACT_BLACKHOLE": "blackhole",
                "FR_ACT_UNREACHABLE": "unreachable",
                "FR_ACT_PROHIBIT": "prohibit",
                "FR_ACT_NAT": "nat",
            }.get(rule['action'], rule['action'])

            if rule['action'] == "to_tbl":
                del rule['action']

            if 'protocol' in rule and rule['protocol'] > 0:
                rule['proto'] = rule['protocol']
                del rule['protocol']

            if 'src_len' in rule and rule['src_len'] > 0:
                rule['from'] = ip_network(
                    '{}/{}'.format(rule['src'], rule['src_len'])).with_prefixlen

            for key in ['src', 'src_len', 'tos', 'protocol']:
                if key in rule:
                    del rule[key]

            rules.append(rule)

        return rules

    def apply(self, ignores, do_apply, by_vrrp, vrrp_type, vrrp_name, vrrp_state):
        ignores = parse_rules(ignores, implicit_defaults=False)

        # get rules that are currently in the kernels list
        krules = self.kernel_rules()
        for rule in self.rules:
            log_str = '#{}'.format(rule['priority'])
            if self.netns.netns != None:
                log_str += "[netns={}]".format(self.netns.netns)

            found = False
            matched = vrrp_match(rule, by_vrrp, vrrp_type, vrrp_name, vrrp_state)
            for i, krule in enumerate(krules):
                if rule_matches(
                    rule,
                    krule,
                    # skip helper attrs like _vrrp
                    [key for key in rule.keys() if key[0] != '_']
                ):
                    found = True

                    # remove kernel routes due to vrrp_match
                    if matched != VRRP_MATCH_DISABLE:
                        del krules[i]

                    break

            if matched not in [VRRP_MATCH_IGNORE, VRRP_MATCH_DISABLE]:
                if found:
                    logger.log_ok(log_str)
                else:
                    logger.log_add(log_str)

                    logger.debug("ip rule add: {}".format(
                        " ".join("{}={}".format(k, v) for k, v in rule.items())))
                    try:
                        if do_apply:
                            self.netns.ipr.rule('add', **rule)
                    except Exception as err:
                        if not isinstance(err, netlinkerror_classes):
                            raise
                        logger.warning('rule setup failed: {}'.format(err.args[1]))

        for rule in krules:
            ignore = False
            for irule in ignores:
                if 'proto' in irule:
                    irule['protocol'] = irule['proto']
                    del irule['proto']
                if rule_matches(rule, irule, irule.keys()):
                    ignore = True
                    break
            if ignore:
                continue

            logger.log_del('#{}'.format(rule['priority']))
            try:
                if do_apply:
                    self.netns.ipr.rule('del', **rule)
            except Exception as err:
                if not isinstance(err, netlinkerror_classes):
                    raise
                logger.warning('removing rule failed: {}'.format(err.args[1]))



from libifstate.util import logger, IfStateLogging
from libifstate.exception import netlinkerror_classes
from ipaddress import ip_interface
from pyroute2.netlink.rtnl.ifaddrmsg import IFA_F_DADFAILED, IFA_F_PERMANENT


class Addresses():
    def __init__(self, netns, iface, addresses):
        self.netns = netns
        self.iface = iface
        self.addresses = []
        for address in addresses:
            self.addresses.append(ip_interface(address))

    def apply(self, ignore, ign_dynamic, do_apply):
        logger.debug('getting addresses', extra={'iface': self.iface, 'netns': self.netns})

        # get ifindex
        idx = next(iter(self.netns.ipr.link_lookup(ifname=self.iface)), None)

        # check if interface exists
        if idx == None:
            return

        # get active ip addresses
        ipr_addr = {}
        addr_add = []
        addr_renew = []
        for addr in self.netns.ipr.get_addr(index=idx):
            flags = addr.get_attr('IFA_FLAGS', 0)
            ip = ip_interface(addr.get_attr('IFA_ADDRESS') +
                              '/' + str(addr['prefixlen']))
            if flags & IFA_F_DADFAILED == IFA_F_DADFAILED:
                logger.debug('{} has failed dad'.format(ip), extra={'iface': self.iface, 'netns': self.netns})
                addr_renew.append(ip)
            ipr_addr[ip] = addr

        for addr in self.addresses:
            if addr in ipr_addr and addr not in addr_renew:
                logger.log_ok('addresses', '= {}'.format(addr.with_prefixlen))
                del ipr_addr[addr]
            else:
                addr_add.append(addr)
                for ip in ipr_addr.keys():
                    if addr.ip == ip.ip:
                        addr_renew.append(ip)
                        break

        for ip, addr in ipr_addr.items():
            if ip in addr_renew or not any(ip in net for net in ignore):
                if not ign_dynamic or ipr_addr[ip]['flags'] & IFA_F_PERMANENT == IFA_F_PERMANENT:
                    logger.log_del('addresses', '- {}'.format(ip.with_prefixlen))
                    try:
                        if do_apply:
                            self.netns.ipr.addr("del", index=idx, address=str(
                                ip.ip), mask=ip.network.prefixlen)
                    except Exception as err:
                        if not isinstance(err, netlinkerror_classes):
                            raise
                        logger.warning('removing ip {}/{} failed: {}'.format(
                            str(ip.ip), ip.network.prefixlen, err.args[1]))

        for addr in addr_add:
            logger.log_change('addresses', '+ {}'.format(addr.with_prefixlen))
            if do_apply:
                try:
                    self.netns.ipr.addr("add", index=idx, address=str(
                        addr.ip), mask=addr.network.prefixlen)
                except Exception as err:
                    if not isinstance(err, netlinkerror_classes):
                        raise
                    logger.warning('adding ip {}/{} failed: {}'.format(
                        str(addr.ip), addr.network.prefixlen, err.args[1]))



from libifstate.util import logger
from libifstate.link.base import Link
from libifstate.exception import LinkCannotAdd

class PhysicalLink(Link):
    def __init__(self, ifstate, netns, name, link, ethtool, vrrp, brport):
        super().__init__(ifstate, netns, name, link, ethtool, vrrp, brport)
        self.cap_create = False
        self.cap_ethtool = True
        self.ethtool = ethtool
 
    def create(self, do_apply, sysctl, excpts, oper="add"):
        logger.warning('unable to create physical link', extra={'iface': self.settings.get('ifname'), 'netns': self.netns})



#!/usr/bin/env python3

from libifstate.parser import YamlParser
from libifstate import __version__, IfState
from libifstate.exception import FeatureMissingError, LinkNoConfigFound, LinkCircularLinked, NetNSNotRoot, ParserValidationError, ParserOpenError, ParserParseError, ParserIncludeError
from libifstate.util import logger, IfStateLogging
from ifstate.vrrp import KEEPALIVED_STATES, KEEPALIVED_IGNORE_STATES
from setproctitle import setproctitle

import argparse
import glob
import logging
import os
from pathlib import Path
import re
import signal
import sys
import yaml


class Actions():
    CHECK = "check"
    APPLY = "apply"
    SHOW = "show"
    SHOWALL = "showall"
    VRRP = "vrrp"
    VRRP_FIFO = "vrrp-fifo"
    VRRP_WORKER = "vrrp-worker"
    SHELL = "shell"

ACTIONS_HELP = {
    "CHECK"      : "dry run update the network config",
    "APPLY"      : "update the network config",
    "SHOW"       : "show running network config",
    "SHOWALL"    : "show running network config (more settings)",
    "VRRP"       : "run as keepalived notify script",
    "VRRP_FIFO"  : "run as keepalived notify_fifo_script",
    "VRRP_WORKER": "worker process for vrrp-fifo",
    "SHELL"      : "launch interactive python shell (pyroute2)",
}

class IfsConfigHandler():
    def __init__(self, fn, soft_schema):
        self.fn = fn
        self.soft_schema = soft_schema

        # require to be called from the root netns
        try:
            # assume init runs in the root netns
            if os.readlink('/proc/1/ns/net') != os.readlink(f'/proc/{os.getpid()}/ns/net'):
                logger.error("Must not be run from inside a netns!")
                raise NetNSNotRoot()
        except OSError as ex:
            logger.debug(f'root netns test: {ex}')
            pass

        self.ifs = self.load_config()
        self.cb = None

    def load_config(self):
        try:
            self.parser = YamlParser(self.fn)
        except ParserOpenError as ex:
            logger.error(
                "Config loading from {} failed: {}".format(ex.fn, ex.msg))
            raise ex
        except ParserParseError as ex:
            logger.error("Config parsing failed:\n\n{}".format(str(ex)))
            raise ex
        except ParserIncludeError as ex:
            logger.error(
                "Config include file {} failed: {}".format(ex.fn, ex.msg))
            raise ex

        try:
            ifs = IfState()
            ifs.update(self.parser.config(), self.soft_schema)
            return ifs
        except ParserValidationError as ex:
            logger.error("Config validation failed for {}".format(ex.detail))
            raise ex
        except FeatureMissingError as ex:
            logger.error(
                "Config uses unavailable feature: {}".format(ex.feature))
            raise ex

    def dump_config(self, fn):
        umask = os.umask(0o077)
        try:
            with open(fn, "w") as fh:
                self.parser.dump(fh)
        finally:
            os.umask(umask)        

def shell():
    from ifstate.shell import IfStateConsole
    import pyroute2

    shell = IfStateConsole()
    shell.interact(banner=f"ifstate {__version__}; pyroute2 {pyroute2.__version__}")


def sighup_vrrp_fifo():
    prefix = 'ifstate-vrrp-fifo@'
    pids = {}

    for pid_file in glob.glob('/run/libifstate/vrrp/[1-9]*.pid'):
        pid = Path(pid_file).stem
        if pid.isnumeric():
            try:
                with open(f'/proc/{pid}/cmdline', encoding='utf-8') as fh:
                    if prefix == fh.read(len(prefix)):
                        # send SIGHUP to reload config
                        os.kill(int(pid), signal.SIGHUP)

                        pids[pid] = fh.read(32)
            except OSError:
                pass

    if len(pids):
        logger.info("")
        logger.info("notified vrrp-fifo handlers...")
        for pid, fifo in sorted(pids.items()):
            logger.info(fifo, extra={'option': pid, 'style': IfStateLogging.STYLE_CHG})

def main():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    group.add_argument("-v", "--verbose", action="count", default=0,
                       help="be more verbose")
    group.add_argument("-q", "--quiet", action="store_true",
                       help="be more quiet, print only warnings and errors")
    parser.add_argument("-s", "--soft-schema", action="store_true",
                        help="ignore schema validation errors, expect ifstatecli to trigger internal exceptions")
    parser.add_argument("-c", "--config", type=str,
                        default="/etc/ifstate/config.yml", help="configuration YaML filename")
    subparsers = parser.add_subparsers(
        dest='action', required=True, help="specifies the action to perform")

    action_parsers = {
        a.lower().replace("_", "-"): subparsers.add_parser(a.lower().replace("_", "-"), help=ACTIONS_HELP[a]) for a in dir(Actions) if not a.startswith('_')
    }

    # Parameters for the vrrp action
    action_parsers[Actions.VRRP].add_argument(
        "type", type=str.lower, choices=["group", "instance"], help="type of vrrp notification")
    action_parsers[Actions.VRRP].add_argument(
        "name", type=str, help="name of the vrrp group or instance")
    action_parsers[Actions.VRRP].add_argument(
        "state", type=str.lower, choices=KEEPALIVED_STATES, help="the new state for the vrrp group or instance")

    # Parameters for the vrrp-fifo action
    action_parsers[Actions.VRRP_FIFO].add_argument(
        "fifo", type=str, help="named FIFO to read state changes from")

    # Parameters for the vrrp-worker action
    action_parsers[Actions.VRRP_WORKER].add_argument(
        "type", type=str.lower, choices=["group", "instance"], help="type of vrrp notification")
    action_parsers[Actions.VRRP_WORKER].add_argument(
        "name", type=str, help="name of the vrrp group or instance")

    args = parser.parse_args()
    if args.verbose > 0:
        lvl = logging.DEBUG
    elif args.quiet:
        lvl = logging.WARNING
    else:
        lvl = logging.INFO

    if args.action == Actions.VRRP_FIFO:
        setproctitle("ifstate-{}@{}".format(args.action, args.fifo))
    else:
        setproctitle("ifstate-{}".format(args.action))

    if args.action == Actions.SHELL:
        shell()
        exit(0)

    ifslog = IfStateLogging(lvl, args.verbose > 1, action=args.action)

    if args.action in [Actions.SHOW, Actions.SHOWALL]:
        # preserve dict order on python 3.7+
        if sys.version_info >= (3, 7):
            yaml.add_representer(
                dict, lambda self, data: yaml.representer.SafeRepresenter.represent_dict(self, data.items()))
        ifs = IfState()
        print(yaml.dump(ifs.show(args.action == Actions.SHOWALL)))

        ifslog.quit()
        exit(0)

    if args.action in [Actions.CHECK, Actions.APPLY, Actions.VRRP, Actions.VRRP_FIFO, Actions.VRRP_WORKER]:
        try:
            ifs_config = IfsConfigHandler(args.config, args.soft_schema)
        except (ParserOpenError,
                ParserParseError,
                ParserIncludeError,
                ParserValidationError,
                FeatureMissingError,
                NetNSNotRoot) as ex:
            ifslog.quit()
            exit(ex.exit_code())

        if args.action == Actions.CHECK:
            try:
                ifs_config.ifs.check()
            except LinkNoConfigFound:
                pass
            except LinkCircularLinked as ex:
                ifslog.quit()
                exit(ex.exit_code())
        elif args.action == Actions.VRRP_FIFO:
            from ifstate.vrrp import vrrp_fifo
            vrrp_fifo(args.fifo, ifs_config)
        elif args.action == Actions.VRRP_WORKER:
            from ifstate.vrrp import vrrp_worker
            vrrp_worker(args.type, args.name, ifs_config)
        else:
            # ignore some well-known signals to prevent interruptions (i.e. due to ssh connection loss)
            signal.signal(signal.SIGHUP, signal.SIG_IGN)
            signal.signal(signal.SIGPIPE, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            try:
                if args.action == Actions.APPLY:
                    ifs_config.ifs.apply()
                    sighup_vrrp_fifo()
                elif args.action == Actions.VRRP:
                    if not args.state in KEEPALIVED_IGNORE_STATES:
                        ifs_config.ifs.apply(
                            args.type, args.name, args.state)
            except LinkNoConfigFound:
                pass
            except LinkCircularLinked as ex:
                ifslog.quit()
                exit(ex.exit_code())

        ifslog.quit()


if __name__ == "__main__":
    main()
